import logging
import traceback


class DuplicateFilter(logging.Filter):
    """
    Logging filter to prevent long-running errors from flooding the log.
    When set, repeated log messages are blocked.
    This will not block alternating messages, and is module-specific.
    """

    def filter(self, record):
        current_log = (record.module, record.levelno, record.msg)
        if current_log != getattr(self, 'last_log', None):
            self.last_log = current_log
            return True
        return False


def brewblox_logger(name: str, dedupe=False):
    """
    Convenience function for creating a module-specific logger.
    """
    logger = logging.getLogger('...'+name[-27:] if len(name) > 30 else name)
    if dedupe:
        logger.addFilter(DuplicateFilter())
    return logger


def strex(ex: Exception, tb=False):
    """
    Generic formatter for exceptions.
    A formatted traceback is included if `tb=True`.
    """
    msg = f'{type(ex).__name__}({str(ex)})'
    if tb:
        trace = ''.join(traceback.format_exception(None, ex, ex.__traceback__))
        return f'{msg}\n\n{trace}'
    else:
        return msg



"""
Offers publishing and subscribing to MQTT events.

Example use:

    import json
    from brewblox_service import mqtt, scheduler

    scheduler.setup(app)
    mqtt.setup(app)

    async def on_message(topic, body):
        print(f'Message topic: {topic}')
        print(f'Message content: {body}')

    mqtt.listen(app, 'brewcast/state/a', on_message)
    mqtt.listen(app, 'brewcast/state/b', on_message)

    mqtt.subscribe(app, 'brewcast/state/#')

    await mqtt.publish('app', 'brewcast/state', json.dumps({'example': True}))
    await mqtt.publish('app', 'brewcast/state/a', json.dumps({'example': True}))
"""

import asyncio
from contextlib import suppress
from dataclasses import dataclass, field
from ssl import CERT_NONE
from typing import Awaitable, Callable, Optional

from aiohttp import web
from aiomqtt import Client, Message, MqttError, TLSParameters, Will
from aiomqtt.types import PayloadType

from brewblox_service import brewblox_logger, features, models, repeater, strex

LOGGER = brewblox_logger(__name__)
MQTT_LOGGER = brewblox_logger('aiomqtt')

ListenerCallback_ = Callable[[str, str], Awaitable[None]]

RECONNECT_INTERVAL_S = 2
INTERACTION_TIMEOUT_S = 5
DEFAULT_PORTS = {
    'mqtt': 1883,
    'mqtts': 8883,
    'ws': 80,
    'wss': 443,
}


def decoded(msg: PayloadType):
    if isinstance(msg, (bytes, bytearray)):
        return msg.decode()
    return msg


@dataclass
class MQTTConfig:
    protocol: str
    host: str
    port: int
    path: str
    client_will: Optional[Will] = None
    transport: str = field(init=False)
    tls_params: Optional[TLSParameters] = field(init=False)

    def __post_init__(self):
        if self.protocol not in ['ws', 'wss', 'mqtt', 'mqtts']:
            raise ValueError(f'Invalid protocol: {self.protocol}')

        if self.protocol.startswith('ws'):
            self.transport = 'websockets'
            self.path = self.path or ''
        else:
            self.transport = 'tcp'
            self.path = ''

        if not self.port:
            self.port = DEFAULT_PORTS[self.protocol]

        if self.protocol in ['mqtts', 'wss']:
            self.tls_params = TLSParameters(cert_reqs=CERT_NONE)
        else:
            self.tls_params = None

    def __str__(self):
        return f'{self.protocol}://{self.host}:{self.port}{self.path}'

    def make_client(self) -> Client:
        client = Client(hostname=self.host,
                        port=self.port,
                        transport=self.transport,
                        websocket_path=self.path,
                        tls_params=self.tls_params,
                        will=self.client_will,
                        logger=MQTT_LOGGER)

        if self.tls_params:
            client._client.tls_insecure_set(True)

        return client


class EventHandler(repeater.RepeaterFeature):
    """
    Connection handler class for MQTT events.
    Handles both TCP and Websocket connections.

    Automatically reconnects and resubscribes when connection is lost.

    Subscribe and listen are handled separately.
    You can subscribe to a wildcard topic, and then set callbacks for specific topics.

    Options are supplied either kwargs, or the service config.

    Kwargs:
        protocol ('ws' | 'wss' | 'mqtt' | 'mqtts'):
            Transport protocol used:
            - ws: websockets
            - wss: websockets over HTTPS
            - mqtt: TCP
            - mqtts: TCP + TLS/SSL

        host (str):
            Hostname of the broker.
            The default hostname of the RabbitMQ broker is 'eventbus'.
            This is accessible only from inside a Brewblox network.
            Outside of a Brewblox network, it is the hostname of the Traefik gateway.

        port (int):
            Port used by the broker.
            The default ports are:
            - ws: 80
            - wss': 443
            - mqtt': 1883
            - mqtts: 8883

            Note that this assumes ws/wss are connected to the Traefik gateway.
            If you are using websockets, and are directly connected to RabbitMQ, values are:
            - ws: 15675
            - wss: 15673

        path (str):
            Path to the broker.
            This value is ignored if mqtt/mqtts protocols are set.
            The default path for the broker is '/eventbus'

            Examples: (formatted as <protocol>://<host>:<port><path>)
                ws://eventbus:15675/eventbus
                wss://BREWBLOX_HOST:443/eventbus
    """

    def __init__(self,
                 app: web.Application,
                 protocol: Optional[models.MqttProtocol] = None,
                 host: Optional[str] = None,
                 port: Optional[int] = None,
                 path: Optional[str] = None,
                 client_will: Optional[Will] = None,
                 publish_will_before_shutdown: bool = True,
                 **kwargs):
        super().__init__(app, **kwargs)

        config: models.BaseServiceConfig = app['config']
        self.config = MQTTConfig(protocol or config.mqtt_protocol,
                                 host or config.mqtt_host,
                                 port or config.mqtt_port,
                                 path or config.mqtt_path,
                                 client_will)
        self.client: Client = self.config.make_client()

        self._ready_ev = asyncio.Event()
        self._connect_delay: int = 0
        self._subscribed_topics: list[str] = []
        self._listeners: list[tuple[str, ListenerCallback_]] = []
        self._publish_will_before_shutdown = publish_will_before_shutdown

    def __str__(self):
        return f'<{type(self).__name__} for {self.config}>'

    @property
    def ready(self) -> asyncio.Event:
        return self._ready_ev

    async def _handle_callback(self, cb: ListenerCallback_, message: Message):
        try:
            await cb(str(message.topic), decoded(message.payload))
        except Exception as ex:
            LOGGER.error(f'Exception handling MQTT callback for {message.topic}: {strex(ex)}')

    async def run(self):
        await asyncio.sleep(self._connect_delay)
        self._connect_delay = RECONNECT_INTERVAL_S

        try:
            async with self.client:
                async with self.client.messages() as messages:
                    if self._subscribed_topics:
                        await self.client.subscribe([(t, 0) for t in self._subscribed_topics],
                                                    timeout=INTERACTION_TIMEOUT_S)

                    LOGGER.debug(f'{self} is ready')
                    self._ready_ev.set()

                    async for message in messages:  # pragma: no cover
                        matching = [cb
                                    for (topic, cb) in self._listeners
                                    if message.topic.matches(topic)
                                    # Workaround for a bug: https://github.com/sbtinstruments/aiomqtt/issues/239
                                    # TODO(Bob) remove when fixed
                                    or (topic.endswith('/#') and message.topic.matches(topic[:-2]))]

                        for cb in matching:
                            asyncio.create_task(self._handle_callback(cb, message))

                        if not matching:
                            LOGGER.debug(f'{self} recv {message}')

        finally:
            self._ready_ev.clear()

    async def before_shutdown(self, app: web.Application):
        if self._publish_will_before_shutdown:
            with suppress(Exception):
                await self.client.publish(**vars(self.config.client_will))

    async def publish(self,
                      topic: str,
                      payload: PayloadType,
                      qos=0,
                      retain=False,
                      err=True,
                      **kwargs):
        try:
            await self.client.publish(topic,
                                      payload,
                                      qos=qos,
                                      retain=retain,
                                      timeout=INTERACTION_TIMEOUT_S,
                                      **kwargs)
            LOGGER.debug(f'publish({topic}) -> OK')
        except MqttError as ex:
            LOGGER.debug(f'publish({topic}) -> {strex(ex)}')
            if err:
                raise ConnectionError(f'Publish error="{strex(ex)}", topic="{topic}"') from ex

    async def subscribe(self, topic: str):
        LOGGER.debug(f'subscribe({topic})')
        self._subscribed_topics.append(topic)
        with suppress(MqttError):
            await self.client.subscribe(topic, timeout=INTERACTION_TIMEOUT_S)

    async def listen(self, topic: str, callback: ListenerCallback_):
        LOGGER.debug(f'listen({topic})')
        self._listeners.append((topic, callback))

    async def unsubscribe(self, topic: str):
        LOGGER.debug(f'unsubscribe({topic})')
        with suppress(MqttError):
            await self.client.unsubscribe(topic, timeout=INTERACTION_TIMEOUT_S)
        with suppress(ValueError):
            self._subscribed_topics.remove(topic)

    async def unlisten(self, topic: str, callback: ListenerCallback_):
        LOGGER.debug(f'unlisten({topic})')
        with suppress(ValueError):
            self._listeners.remove((topic, callback))


def setup(app: web.Application,
          protocol: Optional[models.MqttProtocol] = None,
          host: Optional[str] = None,
          port: Optional[int] = None,
          path: Optional[str] = None,
          client_will: Optional[Will] = None,
          publish_will_before_shutdown: bool = True,
          **kwargs):
    """
    Initializes the EventHandler in the app context.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        protocol (models.MqttProtocol, optional):
            Override the connection protocol.
            If not set, app['config']['mqtt_protocol'] is used.

        host (str, optional):
            Override the broker host.
            If not set, app['config']['mqtt_host'] is used.

        port (int, optional):
            Override the broker port.
            If not set, app['config']['mqtt_port'] is used.

        path (str, optional):
            Override the broker path for WS connections.
            If not set, app['config']['mqtt_path'] is used.

        client_will (Will, optional):
            Set Last Will and Testament for the MQTT connection.

        publish_will_before_shutdown (bool, optional):
            If set, the handler will attempt to send `client_will`
            before a normal shutdown.

    """
    features.add(app,
                 EventHandler(app,
                              protocol=protocol,
                              host=host,
                              port=port,
                              path=path,
                              client_will=client_will,
                              publish_will_before_shutdown=publish_will_before_shutdown,
                              **kwargs))


def fget(app: web.Application) -> EventHandler:
    """
    Get registered EventHandler.
    Requires setup(app) to have been called first.

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    return features.get(app, EventHandler)


async def publish(app: web.Application,
                  topic: str,
                  payload: PayloadType,
                  qos=0,
                  retain=False,
                  err=True,
                  **kwargs):
    """
    Publish a new event message.

    Shortcut for `fget(app).subscribe(topic, message)`.
    Requires setup(app) to have been called first.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            The MQTT message topic. Cannot include wildcards.

        payload (str, bytes, None):
            The message payload.

        retain (bool):
            The MQTT retain flag.
            When a new listener subscribes to a topic,
            it is sent the last retained message by the broker.

        qos (int):
            The MQTT quality-of-service flag.
            Must be 0, 1, or 2.

        err (bool):
            Local flag to determine error handling.
            If set to `False`, no exception is raised when the message could not be published.
    """
    await fget(app).publish(topic=topic,
                            payload=payload,
                            qos=qos,
                            retain=retain,
                            err=err,
                            **kwargs)


async def subscribe(app: web.Application, topic: str):
    """
    Subscribe to event messages.
    Requires setup(app) to have been called first.

    In order to get callbacks for events, listen() must also be used.
    You can register multiple listeners for a single subscribed topic.

    See: http://www.steves-internet-guide.com/understanding-mqtt-topics/

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            A filter for message topics.
            Can include the '+' and '#' wildcards.
    """
    await fget(app).subscribe(topic)


async def listen(app: web.Application, topic: str, callback: ListenerCallback_):
    """
    Set a listener for event messages.
    Requires setup(app) to have been called first.

    In order to get callbacks, subscribe() must also be used.
    You can register multiple listeners for a single subscribed topic.

    See: http://www.steves-internet-guide.com/understanding-mqtt-topics/

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            A filter for message topics.
            Can include the '+' and '#' wildcards.

        callback (Callable[[str, str], Awaitable[None]]):
            The callback that will be invoked if a message is received.
            It is expected to be an async function that takes two arguments: topic and payload.

    """
    await fget(app).listen(topic, callback)


async def unsubscribe(app: web.Application, topic: str):
    """
    Unsubscribe to event messages.
    Requires setup(app) to have been called first.

    Removes a subscription that was set by `subscribe(topic)`.
    Does nothing if no subscription can be found.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            Must match the `topic` argument earlier used in `subscribe(topic)`.
    """
    await fget(app).unsubscribe(topic)


async def unlisten(app: web.Application, topic: str, callback: ListenerCallback_):
    """
    Remove a callback for received event messages.
    Requires setup(app) to have been called first.

    Removes a listener that was set by `listen(topic, callback)`.
    Both `topic` and `callback` must match for the listener to be removed.
    Does nothing if no listener can be found found.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            Must match the `topic` argument earlier used in `listen(topic, callback)`.

        callback (Callable[[str, str], Awaitable[None]]):
            Must match the `callback` argument earlier used in `listen(topic, callback)`.
    """
    await fget(app).unlisten(topic, callback)



"""
Default service middlewares. This includes error handling and automatic CORS headers.
"""


from aiohttp import hdrs, web, web_exceptions
from aiohttp.typedefs import Handler

from brewblox_service import brewblox_logger, strex

LOGGER = brewblox_logger(__name__)


def cors_response(request: web.Request, response: web.Response) -> web.Response:
    response.headers['Access-Control-Allow-Origin'] =\
        request.headers.get('Origin', '*')
    response.headers['Access-Control-Allow-Methods'] =\
        request.headers.get('Access-Control-Request-Method', ','.join(hdrs.METH_ALL))
    response.headers['Access-Control-Allow-Headers'] =\
        request.headers.get('Access-Control-Request-Headers', '*')
    response.headers['Access-Control-Allow-Credentials'] = 'true'
    return response


@web.middleware
async def cors_middleware(request: web.Request, handler: Handler) -> web.Response:
    # preflight requests
    if request.method == 'OPTIONS':
        return cors_response(request, web.Response())

    try:
        response = await handler(request)
        return cors_response(request, response)
    except web_exceptions.HTTPError as ex:  # error_middleware converts other exceptions to HTTPError
        cors_response(request, ex)
        raise ex


@web.middleware
async def error_middleware(request: web.Request, handler: Handler) -> web.Response:
    try:
        return await handler(request)
    except web_exceptions.HTTPError:
        raise
    except Exception as ex:
        raise web_exceptions.HTTPInternalServerError(reason=strex(ex))



from typing import Literal, Optional

from pydantic import BaseModel

MqttProtocol = Literal['mqtt', 'mqtts', 'ws', 'wss']


class BaseServiceConfig(BaseModel):
    host: str
    port: int
    name: str
    debug: bool
    mqtt_protocol: MqttProtocol
    mqtt_host: str
    mqtt_port: Optional[int]
    mqtt_path: str
    history_topic: str
    state_topic: str



"""
Generic startup functions for a brewblox application.

Responsible for parsing user configuration, and creating top-level objects.
This module provides the framework to which service implementations can attach their features.

Example:
    parser = service.create_parser('my_service')
    config = service.create_config(parser)
    app = service.create_app(config)

    async def setup():
        feature_one.setup(app)
        feature_two.setup(app)

    service.run_app(app, setup())
"""

import argparse
import logging
# The argumentparser can't fall back to the default sys.argv if sys is not imported
import sys  # noqa
import tempfile
from typing import Awaitable, Optional, TypeVar

from aiohttp import web
from aiohttp_pydantic import oas

from brewblox_service import brewblox_logger, features, middlewares, models

LOGGER = brewblox_logger(__name__)


def _init_logging(args: argparse.Namespace):
    level = logging.DEBUG if args.debug else logging.INFO
    format = '%(asctime)s %(levelname)-8s %(name)-30s  %(message)s'
    datefmt = '%Y/%m/%d %H:%M:%S'

    logging.basicConfig(level=level, format=format, datefmt=datefmt)
    logging.captureWarnings(True)

    if not args.debug:
        logging.getLogger('aiohttp.access').setLevel(logging.WARN)


def create_parser(default_name: str) -> argparse.ArgumentParser:
    """
    Creates the default brewblox_service ArgumentParser.
    Service-agnostic arguments are added.

    The parser allows calling code to add additional arguments before using it in create_app()

    Args:
        default_name (str):
            default value for the --name commandline argument.

    Returns:
        argparse.ArgumentParser: a Python ArgumentParser with defaults set.

    """
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('-H', '--host',
                        help='Host to which the app binds. [%(default)s]',
                        default='0.0.0.0')
    parser.add_argument('-p', '--port',
                        help='Port to which the app binds. [%(default)s]',
                        default=5000,
                        type=int)
    parser.add_argument('-n', '--name',
                        help='Service name. This will be used as prefix for all endpoints. [%(default)s]',
                        default=default_name)
    parser.add_argument('--debug',
                        help='Run the app in debug mode. [%(default)s]',
                        action='store_true')

    group = parser.add_argument_group('MQTT Event handling')
    group.add_argument('--mqtt-protocol',
                       help='Transport protocol used for MQTT events. [%(default)s]',
                       choices=['mqtt', 'mqtts', 'ws', 'wss'],
                       default='mqtt')
    group.add_argument('--mqtt-host',
                       help='Hostname at which the eventbus can be reached [%(default)s]',
                       default='eventbus')
    group.add_argument('--mqtt-port',
                       help='Port at which the eventbus can be reached [%(default)s]',
                       type=int)
    group.add_argument('--mqtt-path',
                       help='Path used for MQTT events. Only applies if a websockets protocol is used. [%(default)s]',
                       default='/eventbus')
    group.add_argument('--history-topic',
                       help='Eventbus exchange to which logged controller state is broadcast. [%(default)s]',
                       default='brewcast/history')
    group.add_argument('--state-topic',
                       help='Eventbus topic to which volatile controller state is broadcast. [%(default)s]',
                       default='brewcast/state')
    return parser


ConfigT = TypeVar('ConfigT')


def create_config(parser: argparse.ArgumentParser,
                  model: type[ConfigT] = models.BaseServiceConfig,
                  raw_args: Optional[list[str]] = None
                  ) -> ConfigT:
    """
    Parses arguments, and generates the Pydantic config object.
    If you added arguments to the parser, you should provide
    a matching Pydantic model that extends `models.BaseServiceConfig`

    Args:
        parser (argparse.ArgumentParser):
            The parser generated by `create_parser()`,
            optionally extended by caller code.

        model (type[ConfigT], optional):
            The Pydantic model that matches `parser`.

        raw_args (list[str], optional):
            If not set, `sys.argv[1:]` will be used.

    Returns:
        ConfigT: `model` instantiated with the parsed arguments.
    """
    args, unknown_args = parser.parse_known_args(raw_args)
    _init_logging(args)

    if unknown_args:
        LOGGER.error(f'Unknown arguments detected: {unknown_args}')

    return model(**vars(args))


def create_app(config: models.BaseServiceConfig) -> web.Application:
    """
    Creates and configures an Aiohttp application.

    Args:
        config (models.BaseServiceConfig):
            Parsed configuration, generated by create_config().
            This is inserted into the application as `app['config']`.

    Returns:
        web.Application: A configured Aiohttp Application object.
            This Application is not yet running.

    """
    app = web.Application(middlewares=[middlewares.cors_middleware,
                                       middlewares.error_middleware])
    app['config'] = config
    oas.setup(app, url_prefix='/api/doc', title_spec=config.name)

    return app


def run_app(app: web.Application,
            setup: Awaitable = None,
            listen_http: bool = True):
    """
    Runs the application in an async context.
    This function will block indefinitely until the application is shut down.

    Args:
        app (web.Application):
            The Aiohttp Application as created by `create_app()`

        setup (Awaitable):
            If you have setup that should happen async before app start,
            you can provide an awaitable here.
            It will be awaited before application startup.

        listen_http (bool):
            Whether to open a port for the REST API.
            Set to False to disable all REST endpoints.
            This can be useful for services that use communication protocols
            other than REST (such as MQTT), or only have active functionality.
    """
    config: models.BaseServiceConfig = app['config']

    async def _factory() -> web.Application:
        if setup is not None:
            await setup

        prefix = '/' + config.name.lstrip('/')
        for resource in app.router.resources():
            resource.add_prefix(prefix)

        LOGGER.info(f'Service name: {config.name}')
        LOGGER.info(f'Service config: {config}')

        for route in app.router.routes():
            LOGGER.debug(f'Endpoint [{route.method}] {route.resource.canonical}')

        for name, impl in app.get(features.FEATURES_KEY, {}).items():
            LOGGER.debug(f'Feature [{name}] {impl}')

        return app

    if listen_http:
        web.run_app(_factory(), host=config.host, port=config.port)
    else:
        # Listen to a dummy UNIX socket
        # The service still runs, but is not bound to any port
        # This is useful for services without a meaningful REST API
        with tempfile.TemporaryDirectory() as tmpdir:
            web.run_app(_factory(), path=f'{tmpdir}/dummy.sock')
